import re
from .game_objects import GameMasterObject

class Move(GameMasterObject):
    def __init__(self, template, gamemaster_entry):
        super().__init__(
            int(gamemaster_entry["templateId"].split("COMBAT_V")[1].split("_")[0]),
            template,
            gamemaster_entry,
            "combatMove"
        )

        self.type = None

def make_move_list(pogodata):
    moves = []
    for entry in pogodata.raw_gamemaster:
        templateid = entry["templateId"]
        if re.search(r"^COMBAT_V\d{4}_MOVE_.*", templateid):
            template = re.sub(r"^COMBAT_V\d{4}_MOVE_", "", templateid)
            move = Move(template, entry)
            move.type = pogodata.get_type(template=move.raw.get("type"))
            move.name = pogodata.get_locale("move_name_" + str(move.id).zfill(4))
            moves.append(move)
    return moves


class GameObject:
    def __init__(self, id_, template):
        self.id = id_
        self.template = template
        self.name = "?"
    
    def __str__(self):
        return self.template

class GameMasterObject(GameObject):
    def __init__(self, id_, template, gamemaster_entry, settings_name):
        super().__init__(id_, template)
        self.raw = gamemaster_entry.get("data", {}).get(settings_name, {})

def __make_gameobject_list(pogodata, enum, locale_key):
    types = []
    for template, id_ in pogodata.get_enum(enum).items():
        type_ = GameObject(id_, template)
        type_.name = pogodata.get_locale(locale_key.format(template=template))
        types.append(type_)
    return types

def make_type_list(pogodata):
    return __make_gameobject_list(pogodata, "HoloPokemonType", "{template}")

def make_item_list(pogodata):
    return __make_gameobject_list(pogodata, "Item", "{template}_name")



import re

from .util import httpget
from .game_objects import make_type_list, make_item_list
from .moves import make_move_list
from .mons import make_mon_list, check_mons

PROTO_URL = "https://raw.githubusercontent.com/Furtif/POGOProtos/master/base/base.proto"
GAMEMASTER_URL = "https://raw.githubusercontent.com/PokeMiners/game_masters/master/latest/latest.json"
LOCALE_URL = "https://raw.githubusercontent.com/PokeMiners/pogo_assets/master/Texts/Latest%20APK/JSON/i18n_{lang}.json"

class PogoData:
    def __init__(self, language="english", icon_url=""):
        self.__cached_enums = {}
        self.icon_prefix = icon_url
        self.reload(language)

    def reload(self, language="english"):
        self.raw_protos = httpget(PROTO_URL).text
        self.raw_gamemaster = httpget(GAMEMASTER_URL).json()

        raw_locale = httpget(LOCALE_URL.format(lang=language.lower())).json()["data"]
        self.locale = {}
        for i in range(0, len(raw_locale), 2):
            self.locale[raw_locale[i]] = raw_locale[i+1]

        self.items = make_item_list(self)
        self.types = make_type_list(self)
        self.moves = make_move_list(self)

        self.mons = make_mon_list(self)
        check_mons(self)

    def __get_object(self, obj_list, args, match_one=True):
        final = []
        for obj in obj_list:
            wanted = True
            for key, value in args.items():
                if obj.__dict__.get(key) != value:
                    wanted = False
            
            if wanted:
                if match_one:
                    return obj
                final.append(obj)
        
        if not match_one:
            return final

        return None

    def get_mon(self, get_one=True, **args):
        mon = self.__get_object(self.mons, args, get_one)
        if args.get("costume", 0) > 0:
            mon = mon.copy()
            mon.costume = args["costume"]
            mon._gen_asset()

        return mon
    
    def get_type(self, **args):
        return self.__get_object(self.types, args)

    def get_item(self, **args):
        return self.__get_object(self.items, args)

    def get_move(self, **args):
        return self.__get_object(self.moves, args)

    def get_locale(self, key):
        return self.locale.get(key.lower(), "?")

    def get_enum(self, enum, reverse=False):
        cached = self.__cached_enums.get(enum.lower())
        if cached:
            return cached

        proto = re.findall(f"enum {enum} "+r"{[^}]*}", self.raw_protos, re.IGNORECASE)
        if len(proto) == 0:
            raise KeyError(f"Could not find Enum {enum}")
        
        proto = proto[0].replace("\t", "")

        final = {}
        proto = proto.split("{\n")[1].split("\n}")[0]
        for entry in proto.split("\n"):
            k = entry.split(" =")[0]
            v = int(entry.split("= ")[1].split(";")[0])
            final[k] = v
        
        self.__cached_enums[enum.lower()] = final

        if reverse:
            final = {value:key for key, value in final.items()}

        return final



import requests

def httpget(url):
    return requests.get(url)

def gen_uicon(**args):
    icon = ""
    for key, value in args.items():
        if value > 0:
            icon += key + str(value)
    return icon + ".png"



import re
import copy
from .game_objects import GameObject, GameMasterObject
from .util import gen_uicon

class TempEvolution(GameObject):
    def __init__(self, temp_raw):
        super().__init__(0, temp_raw.get("tempEvoId"))
        self.raw = temp_raw
        self.stats = [v for v in self.raw.get("stats", {}).values()]
        self.types = []

class Mon(GameMasterObject):
    def __init__(self, gamemaster_entry, form_id, template):
        super().__init__(
            int(gamemaster_entry["templateId"].split("V")[1].split("_")[0]),
            template,
            gamemaster_entry,
            "pokemonSettings"
        )

        self.form = form_id
        self.base_template = self.raw.get("pokemonId", "")

        self.quick_moves = []
        self.charge_moves = []
        self.types = []
        self.evolutions = []
        self.temp_evolutions = []
        self.stats = [v for v in self.raw.get("stats", {}).values()]

        self.costume = 0

        self.asset_value = None
        self.asset_suffix = None
        self._gen_asset()

    def copy(self):
        return copy.copy(self)

    @property
    def moves(self):
        return self.quick_moves + self.charge_moves

    def _gen_asset(self):
        self.asset = "pokemon_icon_"
        if self.asset_suffix:
            self.asset += str(self.asset_suffix)
        else:
            self.asset += str(self.id).zfill(3) + "_"
            if self.asset_value:
                self.asset += str(self.asset_value)
            else:
                self.asset += "00"
                if self.costume:
                    self.asset += "_" + str(self.costume).zfill(2)

def make_mon_list(pogodata):
    def __typing(mon, type1ref, type2ref):
        typings = [mon.raw.get(type1ref), mon.raw.get(type2ref)]
        for typing in typings:
            if typing:
                mon.types.append(pogodata.get_type(template=typing))

    mons = []
    forms = pogodata.get_enum("Form")
    megas = pogodata.get_enum("HoloTemporaryEvolutionId")
    for entry in pogodata.raw_gamemaster:
        templateid = entry["templateId"]
        if re.search(r"^V\d{4}_POKEMON_.*", templateid):
            template = re.sub(r"^V\d{4}_POKEMON_", "", templateid)
            form_id = forms.get(template, 0)
            mon = Mon(entry, form_id, template)

            locale_key = "pokemon_name_" + str(mon.id).zfill(4)
            mon.name = pogodata.get_locale(locale_key)

            mon.quick_moves = [pogodata.get_move(template=t) for t in mon.raw.get("quickMoves", [])]
            mon.charge_moves = [pogodata.get_move(template=t) for t in mon.raw.get("cinematicMoves", [])]

            __typing(mon, "type", "type2")

            for temp_evo in mon.raw.get("tempEvoOverrides", []):
                evo = TempEvolution(temp_evo)
                evo.id = megas.get(evo.template)
                evo.name = pogodata.get_locale(locale_key + "_" + str(evo.id).zfill(4))
                
                __typing(evo, "typeOverride1", "typeOverride2")

                mon.temp_evolutions.append(evo)
                
            mons.append(mon)

    return mons

def check_mons(pogoinfo):
    form_enums = pogoinfo.get_enum("Form")
    for entry in pogoinfo.raw_gamemaster:
        if re.search(r"^FORMS_V\d{4}_POKEMON_.*", entry.get("templateId", "")):
            formsettings = entry.get("data").get("formSettings")
            forms = formsettings.get("forms", [])
            for form in forms:
                mon = pogoinfo.get_mon(template=form.get("form"))
                if not mon:
                    mon = pogoinfo.get_mon(template=formsettings["pokemon"])
                    mon = mon.copy()
                    mon.template = form.get("form")
                    mon.form = form_enums.get(mon.template)
                    pogoinfo.mons.append(mon) 

                asset_value = form.get("assetBundleValue")
                asset_suffix = form.get("assetBundleSuffix")
                if asset_value or asset_suffix:
                    mon.asset_value = asset_value
                    mon.asset_suffix = asset_suffix
                    mon._gen_asset()

    for mon in pogoinfo.mons:
        evolutions = mon.raw.get("evolutionBranch", [])
        for evo_raw in evolutions:
            if "temporaryEvolution" in evo_raw:
                continue
            evo = pogoinfo.get_mon(
                template=evo_raw.get("form", evo_raw["evolution"])
            )
            mon.evolutions.append(evo)